import aiohttp
from aiohttp import web

from ksrpc.connections import BaseConnection
from ksrpc.serializer.pkl_gzip import serialize, deserialize


def process_response(data):
    if data['status'] == 200:
        return data['data']
    return data


class WebSocketConnection(BaseConnection):
    def __init__(self, url: str, username=None, password=None):
        self._url = url
        self._session = None
        self._ws = None
        self._timeout = aiohttp.ClientTimeout(total=60)
        if username and password:
            self._auth = aiohttp.BasicAuth(login=username, password=password, encoding="utf-8")
        else:
            self._auth = None

    async def __aenter__(self):
        """异步async with"""
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        """异步async with"""
        if self._session:
            await self._session.__aexit__(exc_type, exc_value, traceback)
        if self._ws:
            await self._ws.__aexit__(exc_type, exc_value, traceback)

    async def connect(self):
        if self._session is not None:
            return
        self._session = aiohttp.ClientSession(auth=self._auth, timeout=self._timeout)
        self._ws = await self._session.ws_connect(self._url).__aenter__()

    async def reset(self):
        if self._session is None:
            return
        await self._session.close()
        self._session = None
        self._ws = None

    async def call(self, module, methods, args, kwargs):
        await self.connect()

        d = dict(module=module, methods=methods, args=args, kwargs=kwargs)

        b = serialize(d).read()
        await self._ws.send_bytes(b)
        async for msg in self._ws:
            if msg.type is web.WSMsgType.BINARY:
                return process_response(deserialize(msg.data))
            elif msg.type == web.WSMsgType.ERROR:
                print('ws connection closed with exception %s' % self._ws.exception())
            elif msg.type is web.WSMsgType.CLOSE:
                break

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-05-02

"""
特殊设计的json互转
1. DataFrame和Series的解析最好能在信息完整与空间大小之间平衡，所以选择tight模式
    1. to_dict有tight模式，而to_json却没有
2. 时间希望能正确的转为字符串。pandas._libs.json能正确解析时间

"""
import copy

import numpy as np
import pandas as pd
# 不使用内置json是为了方便处理时间格式
import pandas._libs.json as json


def obj_to_dict(obj):
    """将pandas转成字典，方便后期转成json格式"""
    if isinstance(obj, pd.DataFrame):
        d = obj.to_dict(orient='tight')
    elif isinstance(obj, pd.Series):
        d = obj.to_dict()
    else:
        # 复制一下，防止后期修改了原数据
        if hasattr(obj, 'copy'):
            d = obj.copy()
        else:
            d = copy.copy(obj)
    return d


def dict_to_json(d: dict):
    """字典转json

    调用import pandas._libs.json as json来进行转换，解决时间格式转换问题
    """
    # 复合索引的元组可能简单的只拼接成字符串，导致无法还原
    return json.dumps(d, iso_dates=True)


def json_to_dict(text):
    """json转字典

    调用import pandas._libs.json as json来进行转换，解决时间格式转换问题
    """
    return json.loads(text)


def dict_to_obj(d):
    """dict转pandas"""
    # ndarray直接返回即可
    if isinstance(d, np.ndarray):
        return d
    # list可以转成一维或二维ndarray
    if isinstance(d, list):
        return np.array(d)
    if isinstance(d, dict):
        if 'data' in d:
            # 与to_dict参数对应
            return pd.DataFrame.from_dict(d, orient='tight')
        else:
            # 丢失了index上的name
            return pd.Series(d)


def _display(o1):
    """显示测试"""
    print('=' * 100)
    d1 = obj_to_dict(o1)
    print('-' * 60, 'pandas转dict')
    print(d1)
    j1 = dict_to_json(d1)
    print('-' * 60, 'dict转json')
    print(j1)
    print('-' * 60, 'dict转object')
    print(dict_to_obj(d1))
    d2 = json_to_dict(j1)
    print('-' * 60, 'json转dict')
    print(d2)
    o2 = dict_to_obj(d2)
    print('-' * 60, 'dict转pandas')
    print(o2)


if __name__ == '__main__':
    MINUTES_PER_YEAR = 8
    NUMBER_OF_SHARES = 5

    pd._testing._N = MINUTES_PER_YEAR
    pd._testing._K = NUMBER_OF_SHARES  # 只会显示26个字母

    df = pd._testing.makeTimeDataFrame(freq="5min")
    # df = pd._testing.makeDataFrame()

    # 复合索引测试
    # df = df.set_index('A', append=True)

    _display(df.values)
    _display(df)
    _display(df['B'].values)
    _display(df['B'])
    print(df['B'].to_json())



"""
演示函数
"""
import asyncio
import time

import numpy as np


def sync_say_hi(name):
    time.sleep(1)
    return f'Hello {name}'


async def async_say_hi(name):
    await asyncio.sleep(1)
    return f'Hello {name}'


def div(a, b):
    return a / b


def test():
    import pandas as pd
    df = pd.DataFrame()
    return df


def create_1d_array(target_mb: int = 100):
    """
    创建指定内存大小的一维 NumPy 数组

    参数:
    target_mb -- 目标内存大小 (MB)，默认为 100MB
    dtype -- 数组数据类型，默认为 np.float64

    返回:
    ndarray -- 创建的 NumPy 一维数组
    """
    dtype = np.float64
    # 计算目标字节数
    target_bytes = target_mb * 1024 * 1024

    # 获取数据类型大小
    element_size = np.dtype(dtype).itemsize

    # 计算所需元素数量
    num_elements = target_bytes // element_size

    # 创建随机数组
    arr = np.random.rand(num_elements).astype(dtype)

    # 验证并打印结果
    actual_mb = arr.nbytes / (1024 * 1024)
    print(f"创建成功: {actual_mb:.2f} MB 一维数组")
    print(f"数据类型: {arr.dtype}")
    print(f"元素数量: {arr.size:,}")

    return arr


def test_array():
    arr = create_1d_array()
    return arr[:100]



import aiohttp

from ksrpc.connections import BaseConnection
from ksrpc.serializer.pkl_gzip import deserialize, serialize


async def process_response(r):
    """处理HTTP响应。根据不同的响应分别处理

    Parameters
    ----------
    r: Response

    Returns
    -------
    object
    json
    csv

    """
    if r.status != 200:
        raise Exception(f'{r.status}, {r.text}')
    data = deserialize(await r.content.read())
    if data['status'] == 200:
        return data['data']
    return data


class HttpConnection(BaseConnection):
    """使用aiohttp实现的客户端连接支持同步和异步"""

    def __init__(self, url, username=None, password=None):
        self._url = url
        self._client = None
        self._timeout = aiohttp.ClientTimeout(total=60)
        if username and password:
            self._auth = aiohttp.BasicAuth(login=username, password=password, encoding="utf-8")
        else:
            self._auth = None

    async def __aenter__(self):
        """异步async with"""
        return self

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """异步async with"""
        if self._client:
            await self._client.__aexit__(exc_type, exc_val, exc_tb)

    async def connect(self):
        if self._client is not None:
            return
        self._client = aiohttp.ClientSession(auth=self._auth, timeout=self._timeout)

    async def reset(self):
        if self._client is None:
            return
        await self._client.close()
        self._client = None

    async def call(self, module, methods, args, kwargs):
        """调用函数

        Parameters
        ----------
        module: str
            模块名
        methods: str
            函数全名
        args: tuple
            函数位置参数
        kwargs: dict
            函数命名参数

        """
        await self.connect()

        d = dict(module=module, methods=methods, args=args, kwargs=kwargs)
        files = dict(file=serialize(d).read())
        r = await self._client.post(self._url, data=files)

        return await process_response(r)



# 客户端，用于处理请求
# 需要选择到底是http还是ws，还是nng一类的
import asyncio

from loguru import logger

from ksrpc.connections import BaseConnection


class RpcClient:
    def __init__(self,
                 module: str,
                 connection: BaseConnection,
                 ):
        """初始化

        Parameters
        ----------
        module: str
            模块名
        connection: Connection
            连接对象

        """
        self._module = module
        self._connection = connection
        self._methods = []
        self._lock = asyncio.Lock()

    def __getattr__(self, method):
        # TODO 同一对象在asyncio.gather中使用会混乱
        self._methods.append(method)
        return self

    def __len__(self):
        # 不知怎么回事，被主动调用了
        return 0

    async def __call__(self, *args, **kwargs):
        # 排序，参数顺序统一后，排序生成key便不会浪费了
        kwargs = dict(sorted(kwargs.items()))

        methods_str = '.'.join(self._methods)
        # 用完后得重置，否则第二次用时不正确了
        self._methods = []

        try:
            return await self._connection.call(self._module, methods_str, args, kwargs)
        except asyncio.TimeoutError:
            await self._connection.reset()  # 重置
            logger.warning(f'{self._module}::{methods_str} timeout')
            raise
        except Exception as e:
            await self._connection.reset()  # 重置
            logger.warning(f'{self._module}::{methods_str} error, {e}')
            raise



import multiprocessing
import subprocess
import time

import psutil
import select


def callback(process_name, stream_type, line):
    # 根据流类型添加不同前缀
    if stream_type == "stderr":
        print(f"[ERROR] {process_name}: {line}")
    else:
        print(f"[INFO] {process_name}: {line}")


def run_command(args, callback):
    process_name = multiprocessing.current_process().name

    while True:
        # 子进程崩溃后可重启
        try:
            # 使用 subprocess 运行外部命令
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            print(args, f"启动子进程(PID:{process.pid})")

            # 实时输出日志
            while process.poll() is None:
                # 检查 stdout/stderr 是否有数据
                rlist, _, _ = select.select([process.stdout, process.stderr], [], [], 0.1)
                for stream in rlist:
                    line = stream.readline()
                    if line:
                        # 判断是 stdout 还是 stderr
                        if stream is process.stdout:
                            stream_type = "stdout"
                        elif stream is process.stderr:
                            stream_type = "stderr"
                        else:
                            stream_type = "unknown"

                        callback(process_name, stream_type, line.strip())

            # 等待进程结束
            exit_code = process.wait()
            # 如果正常退出（退出码为0），则跳出循环
            if exit_code == 0:
                print(args, "正常退出")
                break
        except KeyboardInterrupt:
            print(args, "已中断")
            break
        except Exception as e:
            print(args, "运行失败", e)
        finally:
            print(args, "已结束")
            time.sleep(2)


def kill_subprocess(pid):
    # 清除子程序的功能
    parent = psutil.Process(pid)
    for p in parent.children():
        if not p.is_running():
            continue
        print(f"Terminating subprocess: {p.name()} (PID: {p.pid})")
        p.terminate()
        p.wait(timeout=5)

        if not p.is_running():
            continue
        print(f"Force killing subprocess: {p.name()} (PID: {p.pid})")
        p.kill()
        p.wait()


def kill_process(p):
    # 清除程序的功能
    if not p.is_alive():
        return
    print(f"Terminating process: {p.name} (PID: {p.pid})")
    p.terminate()
    p.join(timeout=5)

    if not p.is_alive():
        return
    print(f"Force killing process: {p.name} (PID: {p.pid})")
    p.kill()
    p.join()


class ProcessManager:
    def __init__(self, *processes):
        self.processes = processes

    def __enter__(self):
        for p in self.processes:
            p.start()
            print(f"Started process: {p.name} (PID: {p.pid})")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("Terminating processes...")
        for p in self.processes:
            kill_subprocess(p.pid)
            kill_process(p)
        print("All processes terminated")
        return False  # 不抑制异常



"""
TuShare服务转发
1. 屏蔽set_token功能，由服务端进行统一认证
2. 客户端替换了pro, 服务端将请求转到pro

"""

import tushare as ts

from ..config import TUSHARE_TOKEN

ts.set_token(TUSHARE_TOKEN)
# 用完后清空，防止被客户端获取
del TUSHARE_TOKEN
pro = ts.pro_api()

__path__ = []
__all__ = []


def __getattr__(name):
    return pro.__getattr__(name)


def set_token(token):
    # 防止用户调用到了pro.set_token
    pass


def get_token():
    return ""


def pro_api(token='', timeout=30):
    # 防止用户调用到了pro.pro_api
    pass



import base64

from aiohttp import web
from multidict import MultiDict

from ksrpc.caller import async_call
from ksrpc.config import USER_CREDENTIALS
from ksrpc.serializer.pkl_gzip import deserialize


@web.middleware
async def basic_auth_middleware(request, handler):
    # 检查是否需要跳过认证的路由
    if request.path == '/public':
        return await handler(request)

    # 获取 Authorization 头
    auth_header = request.headers.get('Authorization')

    if auth_header is None or not auth_header.startswith('Basic '):
        return unauthorized_response()

    # 解码凭证
    try:
        auth_bytes = base64.b64decode(auth_header[6:])
        auth_str = auth_bytes.decode('utf-8')
        username, password = auth_str.split(':', 1)
    except (ValueError, UnicodeDecodeError):
        return unauthorized_response()

    # 验证凭证
    if USER_CREDENTIALS.get(username) != password:
        return unauthorized_response()

    # 认证通过，添加用户名到请求对象
    request['user'] = username
    return await handler(request)


def unauthorized_response():
    return web.Response(
        status=401,
        headers={'WWW-Authenticate': 'Basic realm="Restricted Area"'},
        text="Unauthorized"
    )


async def handle(request: web.Request) -> web.StreamResponse:
    form = await request.post()  # 小心大文件
    file = form['file'].file.read()
    del form

    key, buf, data = await async_call(**deserialize(file))
    del file
    del data
    rsp = web.Response(body=buf, headers=MultiDict({'CONTENT-DISPOSITION': f"{key}.pkl.gz"}))
    del buf
    del key
    return rsp


async def websocket_handler(request: web.Request) -> web.StreamResponse:
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    async for msg in ws:
        if msg.type is web.WSMsgType.BINARY:
            key, buf, data = await async_call(**deserialize(msg.data))
            del key
            del data
            await ws.send_bytes(buf)
            del buf
        elif msg.type == web.WSMsgType.ERROR:
            print('ws connection closed with exception %s' % ws.exception())
        elif msg.type is web.WSMsgType.CLOSE:
            break

    return ws


def sync_app(argv):
    # uv run python -m aiohttp.web -H 0.0.0.0 -P 8080 ksrpc.run_app:sync_app
    app = web.Application(middlewares=[
        basic_auth_middleware,  # 注释此行屏蔽Baisc认证
    ])
    app.add_routes([
        web.post("/api/file", handle),
        web.get("/ws/file", websocket_handler),
    ])
    return app


async def async_app():
    # gunicorn ksrpc.run_app:async_app --bind 0.0.0.0:8080 --worker-class aiohttp.GunicornWebWorker
    app = sync_app([])
    return app


if __name__ == '__main__':
    web.run_app(sync_app([]))



import hashlib
import inspect
import sys
from datetime import datetime

from loguru import logger

from ksrpc.serializer.pkl_gzip import serialize

logger.remove()
logger.add(sys.stderr,
           format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level:<8}</level> | PID:{process.id} | TID:{thread.id} | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
           level="INFO", colorize=True)


def make_key(module, methods, args, kwargs):
    """生成缓存key"""
    args_kwargs = f'{module}::{methods}_{repr(args)}_{repr(kwargs)}'
    if len(args_kwargs) > 32:
        args_kwargs = hashlib.md5(args_kwargs.encode('utf-8')).hexdigest()
    return f'{module}::{methods}_{args_kwargs}'


async def async_call(module, methods, args, kwargs):
    """简版API调用。没有各种额外功能"""
    key = make_key(module, methods, args, kwargs)

    # 返回的数据包
    d = dict(status=200,  # status.HTTP_200_OK,
             datetime=datetime.now().isoformat(),  # 加查询时间，缓存中也许可以判断是否过期
             module=module,
             methods=methods,
             args=args,
             kwargs=kwargs)

    try:
        try:
            # 当前server目录下文件，用于特别处理
            api = __import__(f'{__package__}.server.{module}', fromlist=['*'])
        except ModuleNotFoundError as e:
            # 导入系统包
            if module == __package__:
                raise Exception(f'Not Allowed to call {__package__}')
            api = __import__(module, fromlist=['*'])

        # 转成字符串，后面可能于做cache的key
        logger.info(f'{module}::{methods}\t{args}\t{kwargs}'[:200])

        func = api
        for method in methods.split('.'):
            func = getattr(func, method)

        # 可以调用的属性
        if callable(func):
            # 例如os.remove
            type_ = 'function'
            if inspect.iscoroutinefunction(func):
                output = await func(*args, **kwargs)
            else:
                output = func(*args, **kwargs)
        else:
            output = func

        # 结果的类型名
        class_name = output.__class__.__name__

        # 不可调用的直接返回
        if class_name == 'module':
            # 例如os.path，但用法为os.path()
            data = repr(output)
        elif class_name in ('dict_keys', 'dict_values'):
            # 无法序列化，只能强行转换
            # 例如：sys.modules.keys(), 也可以sys.modules.keys.__list()
            data = list(output)
        else:
            # 例如math.pi，但用法math.pi()
            data = output

        d['type'] = type(output).__name__
        d['data'] = data
    except Exception as e:
        d['status'] = 500  # status.HTTP_500_INTERNAL_SERVER_ERROR
        d['type'] = type(e).__name__
        d['data'] = repr(e)

    # 由于错误信息也想缓存，所以这里进行编码
    buf = serialize(d).read()

    return key, buf, d
